#include <iostream>
#include <malloc.h>
#include <objbase.h> //provides the set of constants used in COM applications

using namespace std;

/* GUIDs */

// {C2F7148C-6EA5-41b4-9FB6-B4BBA4C0488F}
const GUID GUID_IFunctionOne =
{ 0xc2f7148c, 0x6ea5, 0x41b4,
{ 0x9f, 0xb6, 0xb4, 0xbb, 0xa4, 0xc0, 0x48, 0x8f } };

// {D7974908-331E-4f7c-BAB8-6711A9D1C0CC}
const GUID GUID_IFunctionTwo =
{ 0xd7974908, 0x331e, 0x4f7c,
{ 0xba, 0xb8, 0x67, 0x11, 0xa9, 0xd1, 0xc0, 0xcc } };

// {83926E15-FDF5-4fd0-B26D-A9F6E428B7A8}
const GUID GUID_IFunctionThree =
{ 0x83926e15, 0xfdf5, 0x4fd0,
{ 0xb2, 0x6d, 0xa9, 0xf6, 0xe4, 0x28, 0xb7, 0xa8 } };

/* the IFunctionOne interface */
interface IFunctionOne: IUnknown
{
	virtual void __stdcall functionOne(void) = 0;
};

/* the IFunctionTwo interface */
interface IFunctionTwo: IUnknown
{
	virtual void __stdcall functionTwo(void) = 0;
};

/* the IFunctionThree interface */
interface IFunctionThree: IUnknown
{
	virtual void __stdcall functionThree(void) = 0;
};

class COMObject: public IFunctionOne, IFunctionTwo, IFunctionThree
{
public:
	/* the COM object constructor */
	COMObject() : reference_counter(0) {}

	/* the COM object destructor */
	~COMObject() {}

private:
	
	/* declare IUnknowns functions */
	virtual HRESULT __stdcall QueryInterface(const GUID &guid, void **storage);
	
	virtual	void __stdcall functionOne(void);
	virtual void __stdcall functionTwo(void);
	virtual void __stdcall functionThree(void);

	virtual ULONG __stdcall AddRef();
	virtual ULONG __stdcall Release();

	/* the reference counter */
	int reference_counter;
};

void COMObject::functionOne(void){}
void COMObject::functionTwo(void){}
void COMObject::functionThree(void){}

HRESULT __stdcall COMObject::QueryInterface(const GUID &guid, void **storage)
{
	/* start by calling the IUnknown base interface with
	   IDD_IUnknown, the ID of the IUnknown interface */
	if(guid == IID_IUnknown)
	{
		/* cast the 'this' pointer to the IUnknown interface
		requested */
		*storage = (IFunctionOne*)this;
	}

	/* the IFunctionOne interface has been requested */
	if(guid == GUID_IFunctionOne)
	{
		/* cast the 'this' pointer to the IFunctionOne
		   interface requested */
		*storage = (IFunctionOne*)this;
	}

	/* the IFunctionTwo interface has been requested */
	else if(guid == GUID_IFunctionTwo)
	{
		/* cast the 'this' pointer to the IFunctionTwo
		   interface requested */
		*storage = (IFunctionTwo*)this;
	}

	/* the IFunctionThree interface has been requested */
	else if(guid == GUID_IFunctionThree)
	{
	/* cast the 'this' pointer to the IFunctionThree
	   interface requested */
	*storage = (IFunctionThree*)this;
	}

	/* the requested interface could not be found */
	else
	{
		*storage = NULL;
		/* QueryInterface should return the error
		   E_NOINTERFACE if it fails */
		return(E_NOINTERFACE);
	}
	/* cast the pointer to IUnknown, subsequently calling the
	   AddRef function so that the reference count can be
	   incremented */
	((IUnknown *)(*storage))->AddRef();
	
	/* QueryInterface should return NOERROR if it succeeds */
	return(NOERROR);
}

ULONG __stdcall COMObject::AddRef()
{
	/* increment the reference count */
	reference_counter = reference_counter + 1;

	/* return this counter for debugging purposes */
	return(reference_counter);
}

ULONG __stdcall COMObject::Release()
{
	/* decrement the reference count */
	reference_counter = reference_counter - 1;
	
	/* check whether the counter isn't 0 */
	if(reference_counter == 0)
	{
		/* the interface pointer is now invalid, so destroy it */
		delete this;
		return 0;
	}
	else
		/* return this counter for debugging purposes */
		return(reference_counter);
}

IUnknown *CoCreateInstance()
{
	/* create the COM object using a pointer to any of our
	   interfaces  used IFunctionOne in this case*/
	IUnknown *comObj = (IFunctionOne *)new(COMObject);
	
	/* increment the reference count */
	comObj->AddRef();
	
	/* return the create COM object */
	return(comObj);
}

int main()
{
	/* the main COM object */
	IUnknown *mainCOMObject = CoCreateInstance();

	IFunctionOne *pInterfaceOne = NULL;
	IFunctionTwo *pInterfaceTwo = NULL;
	IFunctionThree *pInterfaceThree = NULL;

	/* query all three interfaces */
	mainCOMObject->QueryInterface(GUID_IFunctionOne, (void **)&pInterfaceOne);
	mainCOMObject->QueryInterface(GUID_IFunctionTwo, (void **)&pInterfaceOne);
	mainCOMObject->QueryInterface(GUID_IFunctionThree, (void **)&pInterfaceThree);

	/* release the interfaces */
	pInterfaceOne->Release();

	/* release the main COM object */
	mainCOMObject->Release();

	return 0;
}